/*
 * Copyright (c) 2010, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <stdint.h>

#include "config.h"
#include "state.h"
#include "button.h"

/* Number of 'low'-samples needed to detect a press event */
#define BTN_DEBOUNCE_COUNT  5

/* Debounce data structure for one button */
typedef struct {
	unsigned counter  : 7;  // Debounce counter
	unsigned notified : 1;  // Handler notified?
} btn_debounce_t;

/* Debounce data for the two buttons.
 * Making this not as array saves about 60 bytes!
 */
static btn_debounce_t btn_start;
static btn_debounce_t btn_stop;

/* Debounce function, called every 10 ms */
void btn_cyclic_10ms(void)
{
	/* Debouncing for start button */
	if (PINB & (1 << START_BUTTON_PIN)) {
		btn_start.counter = BTN_DEBOUNCE_COUNT;
		btn_start.notified = 0;
	} else {
		if (btn_start.counter > 0)
			btn_start.counter--;
		if ((btn_start.counter == 0) && (!btn_start.notified)) {
			st_enqueue_event(ST_E_START);
			btn_start.notified = 1;
		}
	}

	/* Debouncing for stop button */
	if (PINB & (1 << STOP_BUTTON_PIN)) {
		btn_stop.counter = BTN_DEBOUNCE_COUNT;
		btn_stop.notified = 0;
	} else {
		if (btn_stop.counter > 0)
			btn_stop.counter--;
		if ((btn_stop.counter == 0) && (!btn_stop.notified)) {
			st_enqueue_event(ST_E_STOP);
			btn_stop.notified = 1;
		}
	}
}

/* Initializes button debounce module */
void btn_init(void)
{
	/* Set direction to input */
	DDRB &= ~((1 << START_BUTTON_PIN) | (1 << STOP_BUTTON_PIN));
	/* Enable internal pull-ups */
	PORTB |= ((1 << START_BUTTON_PIN) | (1 << STOP_BUTTON_PIN));

	/* Reset counter info */
	btn_start.counter = BTN_DEBOUNCE_COUNT;
	btn_start.notified = 0;
	btn_stop.counter = BTN_DEBOUNCE_COUNT;
	btn_stop.notified = 0;
}

/* Deinitializes button debounce module */
void btn_deinit(void)
{
	/* Disable internal pull-ups to save power */
	PORTB &= ~((1 << START_BUTTON_PIN) | (1 << STOP_BUTTON_PIN));
}

