/*
 * Copyright (c) 2010, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <avr/io.h>
#include <stdint.h>

#include "config.h"
#include "adc.h"
#include "util.h"
#include "state.h"

#include "cdt.h"

/* Instance of timer data */
static uint16_t cdt_counter;

/* Starts the timer. The external agent is set to 'running'
 * and the indicator (blinker) is started.
 */
void cdt_start(void)
{
	uint32_t factor;

	/* set external agent to 'running' */
	set_pin(CTRL_PIN, CTRL_LEVEL_RUNNING);

	/* get scaling factor specified by control voltage */
#if (VOLT_REVERSE == YES)
	factor = 256 - (uint16_t)adc_getval();
#else
	factor = (uint16_t)adc_getval() + 1;
#endif

	/* set timer to initial value */
	cdt_counter = ((uint32_t)MAX_TIMER_SETTING * factor) / 256;
}

/* Stops the timer. The external agent is set to 'stopped'
 * and the indicator is stopped as well.
 */
void cdt_stop(void)
{
	/* set external agent to 'stopped' */
	set_pin(CTRL_PIN, CTRL_LEVEL_STOPPED);
}

/* Main function of timer, called every second */
void cdt_cyclic_1000ms(void)
{
	if (cdt_counter > 0)
		cdt_counter--;
	if (cdt_counter == 0)
		st_enqueue_event(ST_E_STOP);
}

/* Initializes the CDT application */
void cdt_init(void)
{
	/* set port pin as output */
	DDRB |= (1 << CTRL_PIN);
	/* set pin to defined levels */
	set_pin(CTRL_PIN, CTRL_LEVEL_STOPPED);
}

/* Deinitializes the CDT application */
void cdt_deinit(void)
{
	/* set port pin as input */
	DDRB &= ~(1 << CTRL_PIN);
	/* disable pull-up to save power */
	PORTB &= ~(1 << CTRL_PIN);
}

