/*
 * Copyright (c) 2010, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <stdint.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <stdlib.h>

#include "config.h"
#include "timer.h"
#include "adc.h"
#include "button.h"
#include "cdt.h"
#include "blinker.h"
#include "sleep.h"
#include "util.h"

#include "state.h"

/* States the system can be in */
typedef enum {
	ST_S_SLEEP,
	ST_S_IDLE,
	ST_S_ACTIVE,
	ST_S_COUNT
} st_state_t;
#define ST_S_INVAL ST_S_COUNT

/* Current system state */
static st_state_t st_curstate;

/* Maximum number of enqueued events */
#define ST_QUEUE_LEN 3

/* Counter for enqueued events */
static uint8_t st_event_backlog;
/* Event queue */
static st_event_t st_event_queue[ST_QUEUE_LEN];

/* Prototypes of state initialization functions */
static void st_init_idle(void);
#if (SLEEP_MODE_ENABLED == YES)
static void st_deinit_idle(void);
#endif
static void st_init_active(void);
static void st_deinit_active(void);

/* Type of function pointer to make declaration more readable */
typedef void (*st_initfn_t)(void);

/* State init functions.
 * The columns describe the target state, the rows the source state.
 * The value in the cell is the init function to be called. A NULL
 * is placed where either no init function is needed, or for transitions
 * that are not allowed by the transition table.
 * If the sleep mode is disabled, there is no need for IDLE -> SLEEP.
 */
static st_initfn_t st_init_table[ST_S_COUNT][ST_S_COUNT] PROGMEM = {
	/* from        to SLEEP        to IDLE           to ACTIVE      */
	/* SLEEP  */ { NULL,           st_init_idle,     NULL           },
#if (SLEEP_MODE_ENABLED == YES)
	/* IDLE   */ { st_deinit_idle, NULL,             st_init_active },
#else
	/* IDLE   */ { NULL,           NULL,             st_init_active },
#endif
	/* ACTIVE */ { NULL,           st_deinit_active, NULL           }
};

/* State transition table.
 * The columns identify the received events, the rows the state in which the
 * events are received. The value in the cell is the new state. Invalid
 * transitions are identified by the ST_S_INVAL marker.
 * If the sleep mode is disabled, there is no way back into state SLEEP.
 */
static st_state_t st_trans_table[ST_S_COUNT][ST_E_COUNT] PROGMEM = {
	/* in          got WAKEUP  got START    got STOP    got SLEEPY */
	/* SLEEP  */ { ST_S_IDLE,  ST_S_INVAL,  ST_S_INVAL, ST_S_INVAL },
#if (SLEEP_MODE_ENABLED == YES)
	/* IDLE   */ { ST_S_INVAL, ST_S_ACTIVE, ST_S_INVAL, ST_S_SLEEP },
#else
	/* IDLE   */ { ST_S_INVAL, ST_S_ACTIVE, ST_S_INVAL, ST_S_INVAL },
#endif
	/* ACTIVE */ { ST_S_INVAL, ST_S_INVAL,  ST_S_IDLE,  ST_S_INVAL }
};

/* Initialization for transition from SLEEP to IDLE */
static void st_init_idle(void)
{
	slp_restart();
	tmr_init();
	adc_init();
	btn_init();
	blnk_init();
	cdt_init();

	wdt_enable(WDTO_15MS);
}

#if (SLEEP_MODE_ENABLED == YES)
/* Initialization for transition from IDLE to SLEEP */
static void st_deinit_idle(void)
{
	tmr_deinit();
	adc_deinit();
	btn_deinit();
	blnk_deinit();
	cdt_deinit();

	wdt_disable();
}
#endif

/* Initialization for transition from IDLE to ACTIVE */
static void st_init_active(void)
{
	blnk_start();
	cdt_start();
}

/* Initialization for transition from ACTIVE to IDLE */
static void st_deinit_active(void)
{
	slp_restart();
	blnk_stop();
	cdt_stop();
}

/* Function called by main loop while in state SLEEP */
static void st_main_sleep(void)
{
	slp_powerdown();
}

/* Function called by main loop while in state IDLE */
static void st_main_idle(uint8_t flags)
{
	if (flags & TMR_FLAG_10MS)   btn_cyclic_10ms();
	if (flags & TMR_FLAG_10MS)   adc_cyclic_10ms();
	if (flags & TMR_FLAG_100MS)  slp_cyclic_100ms();
}

/* Function called by main loop while in state ACTIVE */
static void st_main_active(uint8_t flags)
{
	if (flags & TMR_FLAG_10MS)    btn_cyclic_10ms();
	if (flags & TMR_FLAG_10MS)    adc_cyclic_10ms();
	if (flags & TMR_FLAG_100MS)   blnk_cyclic_100ms();
	if (flags & TMR_FLAG_1000MS)  cdt_cyclic_1000ms();
}

/* Dispatch specified event */
static void st_dispatch_event(st_event_t event)
{
	st_state_t new_state;
	st_initfn_t initfn;

	/* Pick new state from transition table */
	new_state = pgm_read_byte_near(&st_trans_table[st_curstate][event]);

	/* If valid, execute init function and switch the state */
	if (new_state != ST_S_INVAL) {
		initfn = (st_initfn_t) pgm_read_word_near(
			&st_init_table[st_curstate][new_state]);
		if (initfn != NULL)
			initfn();
		st_curstate = new_state;
	}
}

/* Main context function called from infinite loop */
void st_main(void)
{
	uint8_t flags;
	uint8_t queue_idx;

	/* query flags from timer module */
	flags = tmr_pop_flags();

	/* Pick the correct main function for current state */
	switch (st_curstate) {
	case ST_S_SLEEP:   st_main_sleep();        break;
	case ST_S_IDLE:    st_main_idle(flags);    break;
	case ST_S_ACTIVE:  st_main_active(flags);  break;
	case ST_S_INVAL:                           break;
	}

	/* Loop over all events enqueued in this cycle */
	for (queue_idx = 0; queue_idx < st_event_backlog; ++queue_idx)
		st_dispatch_event(st_event_queue[queue_idx]);
	st_event_backlog = 0;

	/* Make the watchdog happy */
	wdt_reset();
}

/* Enqueues an event for later dispatching */
void st_enqueue_event(st_event_t event)
{
	TRAP_ASSERT(st_event_backlog < ST_QUEUE_LEN);

	st_event_queue[st_event_backlog++] = event;
}

