/*
 * Copyright (c) 2010, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <avr/io.h>
#include <avr/interrupt.h>

#include "config.h"
#include "util.h"
#include "timer.h"

/* CPU frequency in Hertz */
#define F_CPU  1200000UL

/* Flag register with runnable timers */
static volatile uint8_t tmr_flags;

/* Interrupt Service Routine for compare match A interrupts */
ISR(TIM0_COMPA_vect)
{
	static uint8_t cnt_10ms;
	static uint8_t cnt_100ms;
	static uint8_t cnt_1000ms;

	cnt_10ms++;
	if (cnt_10ms == 10) {
		cnt_10ms = 0;
		cnt_100ms++;
		tmr_flags |= TMR_FLAG_10MS;
	}
	if (cnt_100ms == 10) {
		cnt_100ms = 0;
		cnt_1000ms++;
		tmr_flags |= TMR_FLAG_100MS;
	}
	if (cnt_1000ms == 10) {
		cnt_1000ms = 0;
		tmr_flags |= TMR_FLAG_1000MS;
	}
}

/* Returns the flag register and resets all flags */
uint8_t tmr_pop_flags(void)
{
	uint8_t result;
	uint8_t sreg;

	lock_interrupts(&sreg);
	result = tmr_flags;
	tmr_flags = 0;
	unlock_interrupts(sreg);

	return result;
}

/* Initializes the hardware timer to generate 1 millisecond ticks */
void tmr_init(void)
{
	/* Set timer to CTC mode */
	TCCR0A = (1 << WGM01);
	/* Set prescaler to 8 */
	TCCR0B = (1 << CS01);
	/* Set output compare register for 1ms ticks */
	OCR0A = (F_CPU / 8) / 1000;
	/* Enable output compare A interrupts */
	TIMSK0 = (1 << OCIE0A);
}

/* Deinitializes the hardware timer */
void tmr_deinit(void)
{
	/* Stop timer */
	TCCR0B = 0;
	/* Disable interrupts */
	TIMSK0 = 0;
}

